==> fsh-0.1.tar.gz: fsh-0.1/fshd.py <==
#  fsh - fast remote execution
#  Copyright (C) 1999 by Per Cederqvist.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

import errno
import os
import select
import socket
import string
import sys

import fshlib

class remote:
    def __init__(self, server, method):
	in_fd, self.w = os.pipe()
	self.r, out_fd = os.pipe()
	self.child = os.fork()
	if self.child == 0:
	    # child
	    os.dup2(in_fd, 0)
	    os.dup2(out_fd, 1)
	    os.close(in_fd)
	    os.close(out_fd)
            # FIXME: fix this mess
	    os.execvp(method, [method, server, 'in.fshd'])
	    sys.exit(1)
	# parent
	os.close(in_fd)
	os.close(out_fd)
	self.send_queue = ""
	self.receive_queue = ""
        self.state = 0

    def send(self, data):
	self.send_queue = self.send_queue + data
	if self.w == -1:
	    self.send_queue = ""

    def select_action(self, r, w, e):
	if self.w in w:
	    try:
		sz = os.write(self.w, self.send_queue)
	    except os.error, (eno, emsg):
		if (eno == errno.EINTR or eno == errno.EAGAIN or eno ==
		    errno.EWOULDBLOCK):
		    return
		elif eno == errno.EPIPE:
		    os.close(self.w)
		    self.w = -1
		    print "End of file writing to remote system."
		    return
		else:
		    raise
	    if sz == 0:
		raise "huh? zero-write?"
	    self.send_queue = self.send_queue[sz:]

	if self.r in r:
	    try:
		data = os.read(self.r, 4096)
	    except os.error, (eno, emsg):
		if (eno == errno.EINTR or eno == errno.EAGAIN or eno ==
		    errno.EWOULDBLOCK):
		    return
		else:
		    raise
	    if data == "":
		print "Connection closed by foreign host."
		os.close(self.r)
		self.r = -1
	    else:
		self.receive_queue = self.receive_queue + data

    def select_set(self):
	# Return a tuple of three lists of file descriptors: the read,
	# write and exception fd sets.
	r = []
	w = []
	if self.w != -1 and self.send_queue != "":
	    w.append(self.w)
	if self.r != -1:
	    r.append(self.r)
	return (r, w, [])

    def __readline(self):
        ix = string.find(self.receive_queue, "\n")
        if ix == -1:
            return None
        res = self.receive_queue[:ix]
        self.receive_queue = self.receive_queue[ix+1:]
        return res

    def poll_response(self):
        while self.state == 0:
            s = self.__readline()
            if s == None:
                return [None, None, None]
            if s == "fsh 1":
                self.state = 1
		print "Connection established"
                break
        [cmd, session, data], self.receive_queue = fshlib.parse_line(
            self.receive_queue, 1)

        return [cmd, session, data]

class client:
    def __init__(self, socket):
        self.socket = socket.fileno()
        self.socket_object = socket
        self.send_queue = ""
        self.receive_queue = ""
	self.pending_close = 0

    def select_set(self):
        r = []
        w = []
        e = []
        if self.socket != -1:
            r.append(self.socket)
        if self.socket != -1 and self.send_queue != "":
            w.append(self.socket)
        return r, w, e

    def select_action(self, r, w, e):
        # Handle any possible read/write.
        # Return 0 normally, 1 if the connection was lost.
        if self.socket == -1:
            return 0
        if self.socket in r:
            try:
		data = os.read(self.socket, 4096)
	    except os.error, (eno, emsg):
		if (eno == errno.EINTR or eno == errno.EAGAIN or eno ==
		    errno.EWOULDBLOCK):
		    return 0
		else:
		    raise
	    if data == "":
		self.socket_object.close()
		self.socket = -1
                return 1
	    else:
		self.receive_queue = self.receive_queue + data

        if self.socket in w:
	    try:
		sz = os.write(self.socket, self.send_queue)
	    except os.error, (eno, emsg):
		if (eno == errno.EINTR or eno == errno.EAGAIN or eno ==
		    errno.EWOULDBLOCK):
		    return 0
		elif eno == errno.EPIPE:
		    self.socket_object.close()
		    self.socket = -1
		    return 1
		else:
		    raise
	    if sz == 0:
		raise "huh? zero-write?"
	    self.send_queue = self.send_queue[sz:]
	    if self.pending_close and self.send_queue == "":
		self.socket_object.close()
		self.socket = -1

        return 0
                
    def poll_command(self):
        # return a parsed command, if one is available

        [cmd, data], \
              self.receive_queue = fshlib.parse_line(self.receive_queue, 0)

        return cmd, data

    def send(self, data):
	self.send_queue = self.send_queue + data
	if self.socket == -1:
	    self.send_queue = ""

    def close(self):
	self.pending_close = 1

    def is_closed(self):
	return self.socket == -1


class fshd:
    def __init__(self, server, method):
	# Connect to server using method.  Parse away leading
	# noice until we receive "^fsh ".  Check the version number.
        dir = "/tmp/fshd-%d" % os.getuid()
        try:
            os.mkdir(dir, 0700)
        except os.error, (eno, emsg):
            if eno == errno.EEXIST:
                os.chmod(dir, 0700)
            else:
                raise
	self.r = remote(server, method)
        self.sessions = {}
        self.listen = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.listen.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sockname = dir + "/%s.%s" % (server, method)
	try:
	    os.unlink(sockname)
	except os.error, (eno, emsg):
	    if eno != errno.ENOENT:
		raise
        self.listen.bind(sockname)
        self.listen.listen(3)

    def toploop(self):
        try:
            while 1:
                r, w, e = self.r.select_set()
                r.append(self.listen.fileno())
                for s in self.sessions.keys():
                    r1, w1, e1 = self.sessions[s].select_set()
                    r = r + r1
                    w = w + w1
                    e = e + e1
                r, w, e = select.select(r, w, e, 5)
                self.r.select_action(r, w, e)
                if self.listen.fileno() in r:
                    sock, addr = self.listen.accept()
		    s = 1
		    while (self.sessions.has_key(s) and \
			   not self.sessions[s].is_closed()):
			s = s + 1
                    print s, "start"
                    self.sessions[s] = client(sock)
                while 1:
                    cmd, s, data = self.r.poll_response()
                    if cmd == None:
                        break
                    elif cmd in ["stdout", "stderr", "exit", "signal-exit", \
				 "stdin-flow"]:
			self.sessions[s].send("%s %s\n" % (
			    cmd, fshlib.hollerith(data)))
		    elif cmd in ["eof-stdin", "eof-stdout", "eof-stderr"]:
			self.sessions[s].send("%s\n" % cmd)
		    elif cmd == "eos":
			self.sessions[s].send("eos\n")
			self.sessions[s].close()
			print s, "EOS"
		    else:
			print "Unknown command from in.fshd:", cmd
                for s in self.sessions.keys():
                    if self.sessions[s].select_action(r, w, e) == 1:
                        # fsh died, so kill this session.
                        self.r.send("eos %d\n" % s)
                    while 1:
                        cmd, data = self.sessions[s].poll_command()
                        if cmd == None:
                            break
                        if data == None:
                            self.r.send("%s %d\n" % (cmd, s))
                        else:
                            self.r.send("%s %d %s\n" % (
                                cmd, s, fshlib.hollerith(data)))
			    if cmd == "new":
				print s, "$", data
        except KeyboardInterrupt:
            for s in self.sessions.keys():
                self.sessions[s].send("eos\n")
            sys.exit(0)

def main():
    ac = 1
    method = "ssh"
    if sys.argv[ac] == "-r":
	method = sys.argv[ac+1]
	ac = ac + 2
    server = sys.argv[ac]
    s = fshd(server, method)
    s.toploop()

if __name__ == '__main__':
    main()

==> fsh-0.2.tar.gz: fsh-0.2/fshd.py <==
#  fsh - fast remote execution
#  Copyright (C) 1999 by Per Cederqvist.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

import errno
import getopt
import getpass
import os
import select
import socket
import string
import sys

import fshlib

class remote:
    def __init__(self, method, login, server):
	in_fd, self.w = os.pipe()
	self.r, out_fd = os.pipe()
	self.child = os.fork()
	if self.child == 0:
	    # child
	    cmd = [method]
	    if login != 0:
		cmd = cmd + ["-l", login]
	    cmd = cmd + [server, 'in.fshd']
	    os.dup2(in_fd, 0)
	    os.dup2(out_fd, 1)
	    os.close(in_fd)
	    os.close(out_fd)
	    os.execvp(method, cmd)
	    sys.exit(1)
	# parent
	os.close(in_fd)
	os.close(out_fd)
	self.send_queue = ""
	self.receive_queue = ""
        self.state = 0

    def send(self, data):
	self.send_queue = self.send_queue + data
	if self.w == -1:
	    self.send_queue = ""

    def select_action(self, r, w, e):
	if self.w in w:
	    try:
		sz = os.write(self.w, self.send_queue)
	    except os.error, (eno, emsg):
		if (eno == errno.EINTR or eno == errno.EAGAIN or eno ==
		    errno.EWOULDBLOCK):
		    return
		elif eno == errno.EPIPE:
		    os.close(self.w)
		    self.w = -1
		    print "End of file writing to remote system."
		    return
		else:
		    raise
	    if sz == 0:
		raise "huh? zero-write?"
	    self.send_queue = self.send_queue[sz:]

	if self.r in r:
	    try:
		data = os.read(self.r, 4096)
	    except os.error, (eno, emsg):
		if (eno == errno.EINTR or eno == errno.EAGAIN or eno ==
		    errno.EWOULDBLOCK):
		    return
		else:
		    raise
	    if data == "":
		print "Tunnel closed."
		os.close(self.r)
		self.r = -1
	    else:
		self.receive_queue = self.receive_queue + data

    def select_set(self):
	# Return a tuple of three lists of file descriptors: the read,
	# write and exception fd sets.
	r = []
	w = []
	if self.w != -1 and self.send_queue != "":
	    w.append(self.w)
	if self.r != -1:
	    r.append(self.r)
	return (r, w, [])

    def __readline(self):
        ix = string.find(self.receive_queue, "\n")
        if ix == -1:
            return None
        res = self.receive_queue[:ix]
        self.receive_queue = self.receive_queue[ix+1:]
        return res

    def poll_response(self):
        while self.state == 0:
            s = self.__readline()
            if s == None:
                return [None, None, None]
            if s == "fsh 1":
                self.state = 1
		print "Connection established"
                break
        [cmd, session, data], self.receive_queue = fshlib.parse_line(
            self.receive_queue, 1)

        return [cmd, session, data]

    def is_closed(self):
	return self.r == -1 or self.w == -1

class client:
    def __init__(self, socket):
        self.socket = socket.fileno()
        self.socket_object = socket
        self.send_queue = ""
        self.receive_queue = ""
	self.pending_close = 0

    def select_set(self):
        r = []
        w = []
        e = []
        if self.socket != -1:
            r.append(self.socket)
        if self.socket != -1 and self.send_queue != "":
            w.append(self.socket)
        return r, w, e

    def select_action(self, r, w, e):
        # Handle any possible read/write.
        # Return 0 normally, 1 if the connection was lost.
        if self.socket == -1:
            return 0
        if self.socket in r:
            try:
		data = os.read(self.socket, 4096)
	    except os.error, (eno, emsg):
		if (eno == errno.EINTR or eno == errno.EAGAIN or eno ==
		    errno.EWOULDBLOCK):
		    return 0
		else:
		    raise
	    if data == "":
		self.socket_object.close()
		self.socket = -1
                return 1
	    else:
		self.receive_queue = self.receive_queue + data

        if self.socket in w:
	    try:
		sz = os.write(self.socket, self.send_queue)
	    except os.error, (eno, emsg):
		if (eno == errno.EINTR or eno == errno.EAGAIN or eno ==
		    errno.EWOULDBLOCK):
		    return 0
		elif eno == errno.EPIPE:
		    self.socket_object.close()
		    self.socket = -1
		    return 1
		else:
		    raise
	    if sz == 0:
		raise "huh? zero-write?"
	    self.send_queue = self.send_queue[sz:]
	    if self.pending_close and self.send_queue == "":
		self.socket_object.close()
		self.socket = -1

        return 0
                
    def poll_command(self):
        # return a parsed command, if one is available

        [cmd, data], \
              self.receive_queue = fshlib.parse_line(self.receive_queue, 0)

        return cmd, data

    def send(self, data):
	self.send_queue = self.send_queue + data
	if self.socket == -1:
	    self.send_queue = ""

    def close(self):
	self.pending_close = 1

    def is_closed(self):
	return self.socket == -1

    def drained(self):
	return self.send_queue == ""

class fshd:
    def __init__(self, server, method, login, use_l_flag):
	# Connect to server using method.  Parse away leading
	# noice until we receive "^fsh ".  Check the version number.
        dir = "/tmp/fshd-%d" % os.getuid()
        try:
            os.mkdir(dir, 0700)
        except os.error, (eno, emsg):
            if eno == errno.EEXIST:
                os.chmod(dir, 0700)
            else:
                raise
	self.r = remote(method, use_l_flag and login, server)
        self.sessions = {}
        self.listen = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.listen.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sockname = dir + "/%s.%s.%s" % (server, method, login)
	try:
	    os.unlink(sockname)
	except os.error, (eno, emsg):
	    if eno != errno.ENOENT:
		raise
        self.listen.bind(sockname)
        self.listen.listen(3)

    def toploop(self):
        try:
            while 1:
		drained = self.r.is_closed()
                r, w, e = self.r.select_set()
                r.append(self.listen.fileno())
                for s in self.sessions.keys():
		    if drained and not self.sessions[s].drained():
			drained = 0
                    r1, w1, e1 = self.sessions[s].select_set()
                    r = r + r1
                    w = w + w1
                    e = e + e1
		if drained:
		    sys.exit(0)
                r, w, e = select.select(r, w, e, 5)
                self.r.select_action(r, w, e)
                if self.listen.fileno() in r:
                    sock, addr = self.listen.accept()
		    s = 1
		    while (self.sessions.has_key(s) and \
			   not self.sessions[s].is_closed()):
			s = s + 1
                    print s, "start"
                    self.sessions[s] = client(sock)
                while 1:
                    cmd, s, data = self.r.poll_response()
                    if cmd == None:
                        break
                    elif cmd in ["stdout", "stderr", "exit", "signal-exit", \
				 "stdin-flow"]:
			self.sessions[s].send("%s %s\n" % (
			    cmd, fshlib.hollerith(data)))
		    elif cmd in ["eof-stdin", "eof-stdout", "eof-stderr"]:
			self.sessions[s].send("%s\n" % cmd)
		    elif cmd == "eos":
			self.sessions[s].send("eos\n")
			self.sessions[s].close()
			print s, "EOS"
		    else:
			print "Unknown command from in.fshd:", cmd
                for s in self.sessions.keys():
                    if self.sessions[s].select_action(r, w, e) == 1:
                        # fsh died, so kill this session.
                        self.r.send("eos %d\n" % s)
                    while 1:
                        cmd, data = self.sessions[s].poll_command()
                        if cmd == None:
                            break
                        if data == None:
                            self.r.send("%s %d\n" % (cmd, s))
                        else:
                            self.r.send("%s %d %s\n" % (
                                cmd, s, fshlib.hollerith(data)))
			    if cmd == "new":
				print s, "$", data
        except KeyboardInterrupt:
            for s in self.sessions.keys():
                self.sessions[s].send("eos\n")
            sys.exit(0)

def main():
    ac = 1
    method = "ssh"
    login = None
    use_l_flag = 0
    try:
	opts, args = getopt.getopt(sys.argv[1:], "r:l:")
    except getopt.error, msg:
	sys.stderr.write(msg + "\n")
	sys.exit(1)
    if len(args) != 1:
	sys.stderr.write(
	    "fshd: usage: fshd [ -r rsh ] [ -l login ] host\n")
	sys.exit(1)
    for opt, val in opts:
	if opt == "-r":
	    method = val
	elif opt == "-l":
	    login = val
	    use_l_flag = 1
    if login == None:
	login = getpass.getuser()

    s = fshd(args[0], method, login, use_l_flag)
    s.toploop()

if __name__ == '__main__':
    main()
